#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# pylint: disable=missing-module-docstring

__version__ = '1.2.1'
__author__ = 'Martin Staab, Jean-Baptiste Bayle'
__email__ = 'martin.staab@aei.mpg.de, j2b.bayle@gmail.com'
__copyright__ = \
    '2021, Max Planck Institute for Gravitational Physics ' \
    '(Albert Einstein Institute) and California Institute of Technology'
